import hashlib
import os
import pathlib

import pandas as pd
from clickhouse_driver import Client

from clickhouse_migrations.migrate import apply_migration, migrations_to_apply


class Migrator:
    def __init__(self, db_host: str, db_user: str, db_password: str):
        self.db_host = db_host
        self.db_user = db_user
        self.db_password = db_password

    def connection(self, db_name: str) -> Client:
        return Client(
            self.db_host, user=self.db_user, password=self.db_password, database=db_name
        )

    def create_db(self, db_name):
        with self.connection("") as conn:
            conn.execute(f"CREATE DATABASE IF NOT EXISTS {db_name}")

    @classmethod
    def init_schema(cls, conn):
        conn.execute(
            "CREATE TABLE IF NOT EXISTS schema_versions ("
            "version UInt32, "
            "md5 String, "
            "script String, "
            "created_at DateTime DEFAULT now()"
            ") ENGINE = MergeTree ORDER BY tuple(created_at)"
        )

    def migrate(
        self,
        db_name: str,
        migration_path: pathlib.Path,
        create_db_if_no_exists: bool = True,
    ):
        if create_db_if_no_exists:
            self.create_db(db_name)

        with self.connection(db_name) as conn:
            self.init_schema(conn)

            migrations = [
                {
                    "version": int(f.name.split("_")[0].replace("V", "")),
                    "script": f"{migration_path}/{f.name}",
                    "md5": hashlib.md5(
                        pathlib.Path(f"{migration_path}/{f.name}").read_bytes()
                    ).hexdigest(),
                }
                for f in os.scandir(f"{migration_path}")
                if f.name.endswith(".sql")
            ]
            apply_migration(conn, migrations_to_apply(conn, pd.DataFrame(migrations)))



import pandas as pd


def execute_and_inflate(client, query):
    result = client.execute(query, with_column_types=True)
    column_names = [c[0] for c in result[len(result) - 1]]
    return pd.DataFrame([dict(zip(column_names, d)) for d in result[0]])


def migrations_to_apply(client, incoming):
    current_versions = execute_and_inflate(
        client,
        "SELECT version AS version, script AS c_script, md5 as c_md5 from schema_versions",
    )
    if current_versions.empty:
        return incoming
    if len(incoming) == 0 or len(incoming) < len(current_versions):
        raise AssertionError(
            "Migrations have gone missing, "
            "your code base should not truncate migrations, "
            "use migrations to correct older migrations"
        )

    current_versions = current_versions.astype({"version": "int32"})
    incoming = incoming.astype({"version": "int32"})
    exec_stat = pd.merge(current_versions, incoming, on="version", how="outer")
    committed_and_absconded = exec_stat[
        exec_stat.c_md5.notnull() & exec_stat.md5.isnull()
    ]
    if len(committed_and_absconded) > 0:
        raise AssertionError(
            "Migrations have gone missing, "
            "your code base should not truncate migrations, "
            "use migrations to correct older migrations"
        )

    index = (
        exec_stat.c_md5.notnull()
        & exec_stat.md5.notnull()
        & ~(exec_stat.md5 == exec_stat.c_md5)
    )
    terms_violated = exec_stat[index]
    if len(terms_violated) > 0:
        raise AssertionError(
            "Do not edit migrations once run, "
            "use migrations to correct older migrations"
        )
    return exec_stat[exec_stat.c_md5.isnull()][["version", "script", "md5"]]


def apply_migration(client, migrations):
    if migrations.empty:
        return

    migrations = migrations.sort_values("version")
    for _, row in migrations.iterrows():
        with open(row["script"], "r", encoding="utf-8") as f:
            migration_script = f.read()
            client.execute(migration_script)
            print(
                f"INSERT INTO schema_versions(version, script, md5) "
                f"VALUES({row['version']}, '{row['script']}', '{row['md5']}')"
            )
            client.execute(
                "INSERT INTO schema_versions(version, script, md5) VALUES",
                [
                    {
                        "version": row["version"],
                        "script": row["script"],
                        "md5": row["md5"],
                    }
                ],
            )
